
#include <gsCore/gsTemplateTools.h>

#include  <gsCore/gsAffineFunction.h>
#include  <gsCore/gsAffineFunction.hpp>

namespace gismo {

CLASS_TEMPLATE_INST gsAffineFunction<real_t>;


} // namespace gismo
