/* $NoKeywords: $ */
/*
//
// Copyright (c) 1993-2012 Robert McNeel & Associates. All rights reserved.
// OpenNURBS, Rhinoceros, and Rhino3D are registered trademarks of Robert
// McNeel & Associates.
//
// THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
// ALL IMPLIED WARRANTIES OF FITNESS FOR ANY PARTICULAR PURPOSE AND OF
// MERCHANTABILITY ARE HEREBY DISCLAIMED.
//				
// For complete openNURBS copyright information see <http://www.opennurbs.org>.
//
////////////////////////////////////////////////////////////////
*/


#if !defined(OPENNURBS_VERSION_DEFINITION)
#error Do NOT include opennurbs_version.h in your code.  Use ON::Version() instead.
#endif


// OpenNURBS users:
//   Do not change OPENNURBS_VERSION or the OpenNURBS code
//   that reads 3DM files not work correctly.
#define OPENNURBS_VERSION 201307115

// Subversion revision and branch used for source code.
// Note: ...2012-09-14... is the correct svn branch for
//       the October, 24 release.
#define OPENNURBS_SRC_SVN_REVISION "91919"
#define OPENNURBS_SRC_SVN_BRANCH ".../opennurbs_public/2013-07-11-0000/opennurbs"

// Subversion revision and branch used for documentation.
#define OPENNURBS_DOC_SVN_REVISION "91919"
#define OPENNURBS_DOC_SVN_BRANCH ".../opennurbs_public/2013-07-11-0000/opennurbs"
