#define gsCoDiPack_EXPORTS
#include <gsCoDiPack/gsCoDiPack.h>

namespace codi
{

/* //global data (left here for future reference)
template<> GISMO_EXPORT
ActiveReal<JacobiTape<JacobiTapeTypes<ReverseTapeTypes<double, double, LinearIndexHandler<int> >, ChunkVector> > >::TapeType
ActiveReal<JacobiTape<JacobiTapeTypes<ReverseTapeTypes<double, double, LinearIndexHandler<int> >, ChunkVector> > >::globalTape;

template<> GISMO_EXPORT
ActiveReal<ForwardEvaluation<ForwardTapeTypes<double, double> > >::TapeType
ActiveReal<ForwardEvaluation<ForwardTapeTypes<double, double> > >::globalTape;
*/

}
