#include <gsCore/gsTemplateTools.h>

#include <gsAssembler/gsNewtonCotesRule.h>
#include <gsAssembler/gsNewtonCotesRule.hpp>

namespace gismo
{

    CLASS_TEMPLATE_INST gsNewtonCotesRule<real_t> ;

}
