#include <gsSolver/gsBlockOp.h>
#include <gsSolver/gsBlockOp.hpp>

namespace gismo
{

CLASS_TEMPLATE_INST gsBlockOp<real_t>;
CLASS_TEMPLATE_INST gsBlockOp<std::complex<real_t> >;

} // namespace gismo
