#include <gsCore/gsTemplateTools.h>
#include <gsCore/gsForwardDeclarations.h>
#include <gsCore/gsLinearAlgebra.h>
#include <gsIO/gsXml.h>


#define gsHBoxContainer_EXPORT

#include <gsHSplines/gsHBoxUtils.h>
#include <gsHSplines/gsHBoxContainer.h>
#include <gsHSplines/gsHBoxContainer.hpp>
