/** @file gsCurveLoop_.cpp

    @brief Interface for gsCurveLoop class, representing a loop of curves, in
    anticlockwise order.

    This file is part of the G+Smo library.

    This Source Code Form is subject to the terms of the Mozilla Public
    License, v. 2.0. If a copy of the MPL was not distributed with this
    file, You can obtain one at http://mozilla.org/MPL/2.0/.

    Author(s): A. Falini, A. Mantzaflaris, D.-M. Nguyen, M. Pauley
*/

#include <gsCore/gsTemplateTools.h>
#include <gsModeling/gsCurveLoop.hpp>

namespace gismo
{

CLASS_TEMPLATE_INST gsCurveLoop<real_t> ;

}
