#include <gsCore/gsTemplateTools.h>
#include <gsCore/gsForwardDeclarations.h>
#include <gsCore/gsLinearAlgebra.h>
#include <gsIO/gsXml.h>

#define gsHBox_EXPORT

#include <gsHSplines/gsHBox.h>
#include <gsHSplines/gsHBox.hpp>

#include <gsHSplines/gsHBoxUtils.h>
