#include <gsSolver/gsPatchPreconditionersCreator.h>
#include <gsSolver/gsPatchPreconditionersCreator.hpp>

namespace gismo
{

CLASS_TEMPLATE_INST gsPatchPreconditionersCreator<real_t>;

} // namespace gismo
