#include <gsCore/gsTemplateTools.h>

#include <gsModeling/gsTriMeshToSolid.h>
#include <gsModeling/gsTriMeshToSolid.hpp>

namespace gismo
{

CLASS_TEMPLATE_INST gsTriMeshToSolid<real_t>;

} // namespace gismo
