#include <gsCore/gsTemplateTools.h>

#include <gsAssembler/gsQuadRule.h>
#include <gsAssembler/gsLobattoRule.h>
#include <gsAssembler/gsLobattoRule.hpp>

namespace gismo
{

    CLASS_TEMPLATE_INST gsLobattoRule<real_t> ;

}
