#include <gsCore/gsTemplateTools.h>
#include <gsCore/gsForwardDeclarations.h>
#include <gsCore/gsLinearAlgebra.h>

#define gsAdaptiveMeshing_EXPORT

#include <gsAssembler/gsAdaptiveMeshing.h>
#include <gsAssembler/gsAdaptiveMeshing.hpp>

