#include <gsCore/gsTemplateTools.h>

//Prerequisits
#include <gsUtils/gsL2Projection.h>
#include <gsUtils/gsL2Projection.hpp>
#include <gsAssembler/gsExprAssembler.h>

namespace gismo
{

STRUCT_TEMPLATE_INST gsL2Projection<real_t>;

}
