
#include <gsCore/gsTemplateTools.h>

#include <gsAssembler/gsPoissonAssembler.h>
#include <gsAssembler/gsPoissonAssembler.hpp>

namespace gismo
{
    CLASS_TEMPLATE_INST gsPoissonAssembler<real_t>;
}
